# -*- coding: utf-8 -*-
"""Test radis import.

Start-up time can be monitored with Python 3.7:
https://dev.to/methane/how-to-speed-up-python-application-startup-time-nkf

See for instance:

https://stackoverflow.com/questions/16373510/improving-speed-of-python-module-import

which uses tuna

https://github.com/nschloe/tuna
"""

import radis

assert radis  # silences pyflakes
